/*
  buddy.c
  Buddy Allocator.
  Written by Alex Verstak
  License: use at your own risk.
*/

#include <assert.h>
#include <stdio.h>
#include "buddy.h"
#include "macros.h"
                                       /*definitions*/
#ifndef ALLOCATION_GRANULARITY
#define ALLOCATION_GRANULARITY 0       /*default granularity 4K*/
#endif

#define buddy_shift (12+ALLOCATION_GRANULARITY)
#define buddy_add   ((1<<buddy_shift)-1)
                                       /*dirty internals*/
int   buddy_ptoi(void*);
void *buddy_itop(int,char);
void  buddy_stabilize1(int);
void  buddy_stabilize2(int,char);
void  buddy_trav(int,char);

char * buddy_tree=0;                   /*support structure*/
void * buddy_offset=0;                 /*pool address*/
int    buddy_sz=0;                     /*size in granularity units*/
int    buddy_height=0;                 /*tree height*/
int    buddy_freeblk=0;                /*# free blocks*/

/*
  Get Size of Support Structure.
  Pre:  sz is pool size in bytes, not to exceed 1GB.
  Post: returns needed size in bytes or -1 if bad sz.
  Note: sz=2^(n+buddy_shift).
*/
int buddy_size(int sz)
{
  int s=sz>>buddy_shift;               /*get size in blocks*/
  if(s<=0)                             /*sanity check: >0*/
    return -1;
  while(s>1)                           /*sanity check: power of two*/
    s=s>>1;
  if(s!=1)
    return -1;
  return (2*(sz>>buddy_shift)-1)*sizeof(char);
}                                      /**END buddy_size**/

/*
  Initialize the manager.
  Pre:  t is a char array of sufficient size;
        off is start of the pool;
        sz is pool size in bytes.
  Post: returns 1 if the pool is usable or 0 otherwise.
  Note: see buddy_size.
*/
int buddy_init(char *t,void *off,int sz)
{
  buddy_tree=0; buddy_offset=0;        /*destroy previous info*/
  buddy_sz=0;
  if(buddy_size(sz)<0)                 /*sanity check*/
    return 0;
  if(!t||(int)off+sz<(int)off)         /*sanity check*/
    return 0;
  buddy_sz=sz=sz>>buddy_shift;         /*remember*/
  buddy_tree=t; buddy_offset=off;
  for(buddy_height=1;sz>1;buddy_height++)
    sz=sz>>1;
  buddy_alloc_all();                   /*allocate all memory*/
  return 1;
}                                      /**END buddy_init**/

/*
  Allocate All Memory.
  Pre:  none.
  Post: all memory allocated.
  Note: DESTROYS all information.
*/
void buddy_alloc_all()
{
  int i;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  for(i=0;i<2*buddy_sz-1;i++)
    buddy_tree[i]=-1;
  buddy_freeblk=0;
}                                      /*END buddy_alloc_all**/

/*
  Free All Memory.
  Pre: none.
  Post: all memory freed.
  Note: DESTROYS all information.
*/
void buddy_free_all()
{
  int  i,j,k;
  char level=buddy_height-1;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  for(i=1,j=0;i<=buddy_sz;level--,i=2*i)
    for(k=i;k>0;j++,k--)               /*fill line*/
      buddy_tree[j]=level;
  buddy_freeblk=buddy_sz;
}                                      /**END buddy_free_all**/

/*
  Allocate block.
  Pre:  sz is block size in bytes.
  Post: a pointer is returned, or null if not enough memory.
  Note: sz is rounded up to the power of two.
*/
void *buddy_alloc(int sz)
{
  int  i,found;
  char level=buddy_height-1;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  sz=(sz+buddy_add)>>buddy_shift;      /*sz in blocks now*/
  if(sz<=0||1<<buddy_tree[0]<sz)       /*oops?*/
    return 0;
  for(found=0,i=0;!found&&2*i+2<2*buddy_sz-1;level--)
  {
    assert(buddy_tree[i]<=level&&buddy_tree[i]>=0);
                                       /**left smaller than right?**/
    if(buddy_tree[2*i+1]<=buddy_tree[2*i+2])
      if(1<<buddy_tree[2*i+1]>=sz)     /*...and still large enough?*/
        i=2*i+1;                       /*- go left -*/
      else if(1<<buddy_tree[2*i+2]>=sz)/*at least, right large enough?*/
        i=2*i+2;                       /*- go right -*/
      else
        found=1;                       /*none enough, use current*/
    else                               /**right smaller than left**/
      if(1<<buddy_tree[2*i+2]>=sz)     /*...and still large enough?*/
        i=2*i+2;                       /*- go right -*/
      else if(1<<buddy_tree[2*i+1]>=sz)/*at least, left large enough?*/
        i=2*i+1;                       /*- go left -*/
      else
        found=1;                       /*none enough, use current*/
  }
  assert(1<<buddy_tree[i]>=sz);
  level=buddy_tree[i];
  buddy_tree[i]=-1;                    /*mark allocated*/
  buddy_stabilize1(i);                 /*correct tree*/
  buddy_freeblk-=sz;
  return buddy_itop(i,level);
}                                      /**END buddy_alloc**/

/*
  Free Block.
  Pre:  p points to block.
  Post: returns 1 if block freed or 0 otherwise.
*/
int buddy_free(void *p)
{
  char level;
  int  i=buddy_ptoi(p);
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  if(i<0)
    return 0;                          /*find parent with tree[i]<=0*/
  for(level=0;i>=0&&buddy_tree[i]>=0;level++)
    i=(i-1)/2;
  if(i>=0)                             /*if found, free it*/
  {
    buddy_tree[i]=level;
    buddy_stabilize2(i,level);         /*correct tree*/
    buddy_freeblk+=1<<level;
    return 1;
  }
  else
    return 0;
}                                      /**END buddy_free**/

/*
  Allocate Block At Particular Address.
  Pre:  p is address in the pool, sz is size.
  Post: p is returned, or null if any page busy.
  Note: this function correctly cleans up if error;
        this function is slow: O(n*logN).
*/
void *buddy_alloc_at(void *p,int sz)
{
  void *temp;
  int  tsz;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  sz+=((int)p-(int)buddy_offset)&buddy_add;
  p=(void*)(((int)p-(int)buddy_offset)&~buddy_add);
  tsz=sz=(sz+buddy_add)>>buddy_shift;  /*sz in blocks now*/
  if(sz<=0)
    return 0;
  for(temp=p;sz>0;sz--)                /*allocate sz pages*/
  {
    if(buddy_alloc_page_at(temp)!=temp)
    {
      for(;sz<tsz;sz++)                /*error? clean up*/
      {
        temp=(void*)((int)temp-(1<<buddy_shift));
        buddy_free(temp);
      }
      return 0;
    }
    temp=(void*)((int)temp+(1<<buddy_shift));
  }
  return p;
}                                      /**END buddy_alloc_at**/

/*
  Free Block At Particular Address.
  Pre:  p is address in the pool, sz is size.
  Post: p is returned, or null if any page free.
  Note: this function correctly cleans up if error;
        this function is slow: O(n*logN).
*/
int buddy_free_at(void *p,int sz)
{
  void *temp;
  int  tsz;
  int  i;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  sz+=((int)p-(int)buddy_offset)&buddy_add;
  p=(void*)(((int)p-(int)buddy_offset)&~buddy_add);
  tsz=sz=(sz+buddy_add)>>buddy_shift;  /*sz in blocks now*/
  if(sz<=0)
    return 0;
  for(temp=p;sz>0;sz--)                /*free sz pages*/
  {
    i=buddy_ptoi(temp);
    if(i<0||buddy_tree[i]>=0||!buddy_free(temp))
    {
      for(;sz<tsz;sz++)                /*error? clean up*/
      {
        temp=(void*)((int)temp-(1<<buddy_shift));
        buddy_alloc_page_at(temp);
      }
      return 0;
    }
    temp=(void*)((int)temp+(1<<buddy_shift));
  }
  return 1;
}                                      /**END buddy_free_at**/

/*
  Allocate Page At Particular Address.
  Pre:  p is address in the pool.
  Post: p is returned, or null if page busy.
  Note: use buddy_free to free the page.
*/
void *buddy_alloc_page_at(void *p)
{
  int i;
  assert(buddy_tree&&buddy_sz);        /*stupidity check*/
  p=(void*)(((int)p-(int)buddy_offset)&~buddy_add);
  i=buddy_ptoi(p);                     /*convert p to index*/
  if(i<0)
    return 0;
  if(buddy_tree[i]==0)                 /*free?*/
  {
    buddy_tree[i]=-1;                  /*allocate*/
    buddy_stabilize1(i);
    buddy_freeblk--;
    return p;
  }
  else
    return 0;
}                                      /**END buddy_alloc_page_at**/

/*
  Output Statistics.
  Pre:  none.
  Post: output to stdout.
  Note: it's RECURSIVE.
*/
void buddy_stat()
{
  int  s;
  char level;
  if(!buddy_tree||buddy_sz<=0||buddy_height<=0)
  {
    printf("Buddy system not initialized.\n\n");
    return;
  }
  else
  {
    printf("Buddy system initialized.\n");
    printf("Address %x, %d blocks (%dKB), ",buddy_offset,buddy_sz,buddy_sz<<buddy_shift-10);
    printf("%d blocks free (%dKB).\n",buddy_freeblk,buddy_freeblk<<buddy_shift-10);
    buddy_trav(0,buddy_height-1);
    printf("\n");
  }
}                                      /**END buddy_stat**/

/*
  Internal -- DON'T CALL
  Translates pointer to index in tree.
*/
int buddy_ptoi(void *p)
{
  int i;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  assert((((int)p-(int)buddy_offset)&buddy_add)==0);
  i=((int)p-(int)buddy_offset)>>buddy_shift;
  if(i<0||i>=buddy_sz)                 /*doesn't fit?*/
    return -1;
  else
    return (i+buddy_sz-1);             /*i is an index now*/
}                                      /**END buddy_ptoi**/

/*
  Internal -- DON'T CALL
  Translates index in tree to pointer.
*/
void *buddy_itop(int i,char level)
{
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  assert(i<2*buddy_sz-1&&i>=0);
  assert(level<buddy_height);
  i-=(1<<buddy_height-level-1)-1;      /*cryptic conversion i->address*/
  return (void*)((i<<level+buddy_shift)+(int)buddy_offset);
}                                      /**END buddy_itop**/

/*
  Internal -- DON'T CALL.
  Walks up after allocation: non-coalescing stabilize2.
*/
void buddy_stabilize1(int i)
{
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  assert(i<2*buddy_sz-1&&i>0);
  while(i>0)
  {
    i=(i-1)/2;                         /*go parent*/
    buddy_tree[i]=max(buddy_tree[2*i+1],buddy_tree[2*i+2]);
  }                                    /*parent=greater of children*/
}                                      /**END buddy_stabilize1**/

/*
  Internal -- DON'T CALL.
  Walks up after freing.
*/
void buddy_stabilize2(int i,char level)
{
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  assert(i<2*buddy_sz-1&&i>0);
  assert(level<buddy_height);
  while(i>0)
  {
    i=(i-1)/2;                         /*go parent*/
    assert(buddy_tree[i]<=level+1);
    if(buddy_tree[2*i+1]>buddy_tree[2*i+2])
      buddy_tree[i]=buddy_tree[2*i+1]; /*-use left-*/
    else if(buddy_tree[2*i+1]<buddy_tree[2*i+2])
      buddy_tree[i]=buddy_tree[2*i+2]; /*-use right-*/
    else if(buddy_tree[2*i+1]==level)
      buddy_tree[i]=level+1;           /*coalesce*/
    else
      buddy_tree[i]=buddy_tree[2*i+1]; /*otherwise, use one*/
    level++;                           /*next level*/
  }
}                                      /**END buddy_stabilize2**/

/*
  Internal -- DON'T CALL
  Traverse Tree.
*/
void buddy_trav(int i,char level)
{
  int   block_no=(i-((1<<buddy_height-level-1)-1))*(1<<level);
  void *block_addr=buddy_itop(i,level);
  int   block_sz=1<<level;
  assert(buddy_tree&&buddy_sz>0&&buddy_height>0);
  assert(i<2*buddy_sz-1&&i>=0);
  assert(level<buddy_height);
  if(buddy_tree[i]==level)
    printf("#%6d, at %08x, size %6d, Free\n",block_no,block_addr,block_sz);
  else if(buddy_tree[i]<0)
    printf("#%6d, at %08x, size %6d, Busy\n",block_no,block_addr,block_sz);
  else
  {
    buddy_trav(2*i+1,level-1);
    buddy_trav(2*i+2,level-1);
  }
}

/*-------------- end of buddy.c ----------------*/

